#ifndef GUARD_effect_h
#define GUARD_effect_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <cmath>
#include <mad.h>
#include "preference.h"

namespace Effect {

    class Base {
	public:	
	    Base(): xn1(0.0f), xn2(0.0f), yn1(0.0f), yn2(0.0f) {}
	    ~Base() {}

	    void Reset()
	    {
		xn1 = xn2 = yn1 = yn2 = 0.0f;
	    }
	    
	    float xn1, xn2, yn1, yn2;
    };

    class Stereo {
	public:
	    Stereo(const double& l, const double& r)
	    {
		left = l;
		right = r;
	    }
	    
	    Stereo()
	    {
		left = 0.0;
		right = 0.0;
	    }
	    ~Stereo() {}
	    
	    double left, right;
    };

    class BiquadCoeff {
	public:
	    BiquadCoeff(): omega_(0.0), sn_(0.0), cs_(0.0), a_(0.0), shape_(0.0),
		beta_(0.0), b0_(0.0), b1_(0.0), b2_(0.0), a0_(0.0), a1_(0.0), a2_(0.0),
		alpha_(0.0), norm_(0.0), gain_(0.0f), sampleRate_(0), freq_(0) {}
	    virtual ~BiquadCoeff() {}

	    virtual void Calc(const int&, const int&, const float&, const double&) = 0;

	    void Apply(double& orig, Base& base) const
	    {
		double in = orig;
		double out = a0_;
		out *= in;
		out += a1_ * base.xn1;
		out += a2_ * base.xn2;
		out -= b1_ * base.yn1;
		out -= b2_ * base.yn2;
		    
		base.xn2 = base.xn1;
		base.xn1 = in;
		base.yn2 = base.yn1;
		base.yn1 = out;
		orig = out;    
	    }
	    
	    double Return(double& orig, Base& base) const
	    {
		double in = orig;
		double out = a0_;
		out *= in;
		out += a1_ * base.xn1;
		out += a2_ * base.xn2;
		out -= b1_ * base.yn1;
		out -= b2_ * base.yn2;
		    
		base.xn2 = base.xn1;
		base.xn1 = in;
		base.yn2 = base.yn1;
		base.yn1 = out;

		return out;
	    }

	    void Reset()
	    {
		omega_ = sn_ = cs_ = a_ = shape_ = beta_ = b0_ = b1_ = b2_ = a0_ = 0.0;
		a1_ = a2_ = alpha_ = norm_ = 0.0;
		gain_ = 0.0f;
		sampleRate_ = 0;
		freq_ = 0;
	    }

	    const int& SampleRate() const { return sampleRate_; }
	    const int& Freq() const { return freq_; }

	protected:    
	    double omega_, sn_, cs_, a_, shape_, beta_, b0_, b1_, b2_, a0_, a1_, a2_, alpha_, norm_;
	    float gain_;
	    int sampleRate_;
	    int freq_;
    };

    class LowShelf : public BiquadCoeff {
	public:
	    LowShelf() {}
	    virtual ~LowShelf() {}

	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;
		gain_ = gain;
		shape_ = Q;
		
		omega_ = 2.0;
		omega_ *= M_PI;
		omega_ *= freq_;
		omega_ /= sampleRate_;
		
		sn_ = sin(omega_);
		cs_ = cos(omega_);

		a_ = log(10.0);
		a_ *= gain_;
		a_ *= 0.025;
		a_ = exp(a_);
		
		alpha_ = sn_;
		alpha_ *= 0.5;
		alpha_ /= shape_;
		
		beta_ = sqrt(a_ * a_ + 1.0);
		beta_ /= shape_;
		beta_ -= pow((a_ - 1.0) , 2.0);

		norm_ = a_ + 1.0;
		norm_ += (a_ - 1.0) * cs_;
		norm_ += beta_ * sn_;
	
		double reverseNorm = 1.0 / norm_;
		
		a0_ = a_;
		a0_ += 1.0;
		a0_ -= (a_ - 1.0) * cs_;
		a0_ += beta_ * sn_;
		a0_ *= a_;
		a0_ *= reverseNorm;

		a1_ = a_;
		a1_ -= 1.0;
		a1_ -= (a_ + 1.0) * cs_;
		a1_ *= 2.0;
		a1_ *= a_;
		a1_ *= reverseNorm;
		
		a2_ = a_;
		a2_ += 1.0;
		a2_ -= (a_ - 1.0) * cs_;
		a2_ -= beta_ * sn_;
		a2_ *= a_;
		a2_ *= reverseNorm;

		b1_ = a_;
		b1_ -= 1.0;
		b1_ += (a_ + 1.0) * cs_;
		b1_ *= -2.0;
		b1_ *= reverseNorm;
		
		b2_ = a_;
		b2_ += 1.0;
		b2_ += (a_ - 1.0) * cs_;
		b2_ -= beta_ * sn_;
		b2_ *= reverseNorm;
		
	    }

    };	    

    class HighShelf : public BiquadCoeff
    {
	public:
	    HighShelf() {}
	    virtual ~HighShelf() {}

	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;
		gain_ = gain;
		shape_ = Q;

		omega_ = 2.0;
		omega_ *= M_PI;
		omega_ *= freq_;
		omega_ /= sampleRate_;

		sn_ = sin(omega_);
		cs_ = cos(omega_);

		a_ = log(10.0);
		a_ *= gain_;
		a_ *= 0.025;
		a_ = exp(a_);

		alpha_ = sn_;
		alpha_ /= 2.0 * shape_;
		
		beta_ = sqrt(a_ * a_ + 1.0);
		beta_ /= shape_;
		beta_ -= pow((a_ - 1.0), 2.0);
		
		norm_ = a_;
		norm_ += 1.0;
		norm_ -= (a_ - 1.0) * cs_;
		norm_ += beta_ * sn_;

		double reverseNorm = 1.0 / norm_;

		b1_ = a_ - 1.0;
		b1_ -= (a_ + 1.0) * cs_;
		b1_ *= 2.0;
		b1_ *= reverseNorm;

		b2_ = a_;
		b2_ += 1.0;
		b2_ -= (a_ - 1.0) * cs_;
		b2_ -= beta_ * sn_;
		b2_ *= reverseNorm;

		a0_ = a_;
		a0_ += 1.0;
		a0_ += (a_ - 1.0) * cs_;
		a0_ += beta_ * sn_;
		a0_ *= a_;
		a0_ *= reverseNorm;
		
		a1_ = a_;
		a1_ -= 1.0;
		a1_ += (a_ + 1.0) * cs_;
		a1_ *= -2.0;
		a1_ *= a_;
		a1_ *= reverseNorm;
		
		a2_ = a_;
		a2_ += 1.0;
		a2_ += (a_ - 1.0) * cs_;
		a2_ -= beta_ * sn_;
		a2_ *= a_;
		a2_ *= reverseNorm;
		
	    }
    };

    class Peak : public BiquadCoeff
    {
	public:
	    Peak() {}
	    virtual ~Peak() {}

	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;
		gain_ = gain;
		shape_ = Q;
		
		omega_ = 2.0;
		omega_ *= M_PI;
		omega_ *= freq_;
		omega_ /= sampleRate_;

		sn_ = sin(omega_);
		cs_ = cos(omega_);

		a_ = log(10.0);
		a_ *= gain_;
		a_ *= 0.025;
		a_ = exp(a_);

		alpha_ = sn_;
		alpha_ /= 2.0 * shape_;
		
		beta_ = sqrt(a_ * a_ + 1.0);
		beta_ /= shape_;
		beta_ -= pow((a_ - 1.0), 2.0);
		
		norm_ = alpha_;
		norm_ /= a_;
		norm_ += 1.0;

		double reverseNorm = 1.0 / norm_;

		b1_ = -2.0;
		b1_ *= cs_;
		b1_ *= reverseNorm;
	
		b2_ = 1.0;
		b2_ -= alpha_ / a_;
		b2_ *= reverseNorm;	

		a0_ = 1.0;
		a0_ += alpha_ * a_;
		a0_ *= reverseNorm;

		a1_ = -2.0;
		a1_ *= cs_;
		a1_ *= reverseNorm;

		a2_ = 1.0;
		a2_ -= alpha_ * a_;
		a2_ *= reverseNorm;
	    }
    };

    class LowPass : public BiquadCoeff
    {
	public:
	    LowPass() {}
	    virtual ~LowPass() {}

	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;
		gain_ = gain;

		omega_ = 1.0;
		omega_ /= tan(M_PI * freq_ / sampleRate_);

		a0_ = 1.0;
		a0_ /= (1.0 + gain_ * omega_ + omega_ * omega_);
		
		a1_ = 2.0;
		a1_ *= a0_;

		a2_ = a0_;

		b0_ = 2.0;
		b0_ *= 1.0 - omega_ * omega_;
		b0_ *= a0_;
	
		b1_ = 1.0 - gain_ * omega_ + omega_ * omega_;
		b1_ *= a0_;
			
	    }
    };

    class HighPass : public BiquadCoeff
    {
	public:
	    HighPass() {}
	    virtual ~HighPass() {}

	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;
		gain_ = gain;
		omega_ = tan(M_PI * freq_ / sampleRate_);

		a0_ = 1.0;
		a0_ /= 1.0 + gain_ * omega_ + omega_ * omega_;

		a1_ = -2.0;
		a1_ *= a0_;

		a2_ = a0_;

		b0_ = 2.0;
		b0_ *= omega_ * omega_ - 1.0;
		b0_ *= a0_;

		b1_ = 1.0 - gain_ * omega_ + omega_ * omega_;
		b1_ *= a0_;
	    }
    };

    static Base BassBoost_L[2], BassBoost_R[2];
    static Base BassBoost2_L[3], BassBoost2_R[3];
    static Base TrebleBoost_L, TrebleBoost_R;
    static Base GraphicEQ_L[11], GraphicEQ_R[11];
    static Base Room_L, Room_R;
    static Base Room2_L[6], Room2_R[6];
    static Base Harmony_L, Harmony_R;

    static Peak Bass50, Bass2_80;;
    
    static Peak EQ50, EQ100, EQ200, EQ400, EQ800, EQ1500, EQ3k, EQ5k;
    static Peak EQ7k, EQ10k, EQ15k;
    
    static LowShelf Bass100, Bass2_199;
    static LowShelf Room2_35, Room2_62, Room2_279, Room2_852;
    static HighShelf Treble5k, Bass2_1309;
    static HighShelf Room2_3713, Room2_13545; 
    static LowPass LPF_;
    static HighPass HPF_;

    inline double* prevLeftSample()
    {
	static double prevLeftSample_[5];

	return prevLeftSample_;
    }
    
    inline double* prevRightSample()
    {
	static double prevRightSample_[5];

	return prevRightSample_;
    }

    inline double* prevLeftSample1()
    {
	static double prevLeftSample1_[3];

	return prevLeftSample1_;
    }

    inline double* prevRightSample1()
    {
	static double prevRightSample1_[3];

	return prevRightSample1_;
    }

    inline double* peakSample()
    {
	static double peakSample_[2];

	return peakSample_;
    }

    inline double* prevAvg()
    {
	static double prevAvg_[2];

	return prevAvg_;
    }
    
    inline double& feedGain()
    {
	static double feedGain_;

	return feedGain_;
    }

    inline double& feedRhythme()
    {
	static double feedRhythme_;

	return feedRhythme_;
    }

    inline double& clipAbsInL()
    {
	static double clipAbsInL_;

	return clipAbsInL_;
    }

    inline double& clipAbsInR()
    {
	static double clipAbsInR_;

	return clipAbsInR_;
    }

    inline double& amplifyRatio()
    {
	static double amplifyRatio_;

	return amplifyRatio_;
    }

    inline double& amplifyGain()
    {
	static double amplifyGain_;

	return amplifyGain_;
    }

    inline int& prevCount()
    {
	static int prevCount_;

	return prevCount_;
    }
    
    inline int& prevRoomCount()
    {
	static int prevRoomCount_;

	return prevRoomCount_;
    }

    inline int& prevHarmonyCount()
    {
	static int prevHarmonyCount_;

	return prevHarmonyCount_;
    }
    
    void Reset();

    void Process(const Preference&, const int& , Stereo&);
    void Amplify(double&, double&, const float&);
    
    void Clip(double&, double&);
    
    void BassBoost(const int&, double&, double&);
    void BassBoost2(const int&, double&, double&);
    void TrebleBoost(const int&, double&, double&);
    
    void EQ(BiquadCoeff&, const int&, double&, double&, const int&, const float&, Base&, Base&, const float& = 1.0);
    void GraphicEQ(const int&, double&, double&, const float* const);
    
    Stereo LPF(BiquadCoeff&, const int&, double&, double&, const int&, const double&, Base&, Base&);
    Stereo HPF(BiquadCoeff&, const int&, double&, double&, const int&, const double&, Base&, Base&);
    
    void StereoEnhance(double&, double&, const float&);
    void Rhythme(double&, double&, const double&);
    void Room(const int&, double&, double&, const double&);
    void Room2(const int&, double&, double&, const double&);
    void Harmony(const int&, double&, double&, const double&);
    void Distortion(double&, double&, const double&);

}

#endif
