// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <iostream>
#include <fstream>

#include <map>

#include <gdkmm.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <libintl.h>

#include "../include/playlist.h"
#include "../include/mainwin.h"
#include "../include/tag_mp3.h"
#include "../include/tag_ogg.h"
#include "../include/tag_wma.h"
#include "../include/fileinfodlg.h"
#include "../include/trace.h"
#include "../include/misc.h"

std::auto_ptr<Tag> CreateId3Tag(const std::string& filename)
{
    return std::auto_ptr<Tag>(new MP3Tag(filename));
}

std::auto_ptr<Tag> CreateOGGTag(const std::string& filename)
{
    return std::auto_ptr<Tag>(new OGGTag(filename));
}

std::auto_ptr<Tag> CreateWMATag(const std::string& filename)
{
    return std::auto_ptr<Tag>(new WMATag(filename));
}


Playlist::Playlist(MainWnd& mainWnd, const std::string& listId, const unsigned int& titleWidth)
: listId_(listId),
mainWnd_(mainWnd),
tagParsingThread_(NULL),
count_(0),
tagProcessCount_(0),
inTagParsing_(false),
stopTagParsing_(false)
{
    #warning "Fix this magic constant!"
    tagProcessCount_ = (mainWnd_.prevHeight_ - 170) / 25 + 2;
    
    treeModel_ = Gtk::ListStore::create(columns_);
    set_model(treeModel_);

    append_column(gettext("Number"), columns_.id_);
    append_column(gettext("Title"), columns_.title_);
    append_column(gettext("Time"), columns_.time_);

    Gtk::Menu_Helpers::MenuList& popupMenuList = menuPopup_.items();
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Add files"),
	sigc::mem_fun(mainWnd_, &MainWnd::OnMenuFileAdd) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Add a _directory"),
	sigc::mem_fun(mainWnd_, &MainWnd::OnMenuFileDirAdd) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Remove"),
	sigc::mem_fun(*this, &Playlist::OnMenuRemove) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Re_move all"),
	sigc::mem_fun(*this, &Playlist::OnMenuRemoveAll) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Sort filename"),
	sigc::mem_fun(*this, &Playlist::OnMenuSortFilename) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Information"),
	sigc::mem_fun(*this, &Playlist::OnMenuInfo) ));

    // Column 1 is title.
    get_column(1)->set_resizable(true);
    get_column(1)->set_fixed_width(titleWidth);
    get_column(1)->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
    
    tagFactory_ = std::auto_ptr<TagFactory>(new TagFactory);
    tagFactory_->Register("mp1", &CreateId3Tag);
    tagFactory_->Register("mp2", &CreateId3Tag);
    tagFactory_->Register("mp3", &CreateId3Tag);
    tagFactory_->Register("ogg", &CreateOGGTag);
    tagFactory_->Register("wma", &CreateWMATag);

    LoadList();

    treeModel_->set_sort_func(columns_.filename_, 
    	sigc::mem_fun(*this, &Playlist::FilenameCompare));

    signalRowToFill_.connect(sigc::mem_fun(*this, &Playlist::RowFill));

    Gtk::TreeModel::iterator iter = treeModel_->children().begin();

    if(iter) {
	inTagParsing_ = true;
	tagParsingThread_ = Glib::Thread::create( sigc::bind(
	   sigc::mem_fun(*this, &Playlist::GetInfoFromTag), iter), true );
    }
}

Playlist::~Playlist()
{
    StopThread();
    
    SaveList();
}

void Playlist::Add(const std::string& filename)
{
    Gtk::TreeModel::Row row = *(treeModel_->append());
    row[columns_.tagParsed_] = false;
    row[columns_.id_] = ++count_;
    row[columns_.filename_] = filename;
    row[columns_.timeSeconds_] = 1;

}

std::string Playlist::Get()
{
    return GetFilenameFromTreeiter(pos_);
}

std::string Playlist::GetNext()
{
    if(pos_) {
	
	if( pos_ == --(treeModel_->children().end()) ) {
	   
	    if(!mainWnd_.repeatPlay_) {

		mainWnd_.listState_ = MainWnd::LISTSTATE_LAST;
			
		return std::string();
	    }
	}

	return GetFilenameFromTreeiter(++pos_);	
	
    } else
	return GetFilenameFromTreeiter(pos_);	
}

std::string Playlist::GetPrev()
{
    if(pos_) {

	if(pos_ == treeModel_->children().begin()) {
	    
	    if(mainWnd_.repeatPlay_)
		pos_ = treeModel_->children().end();
	    else {	

		mainWnd_.listState_ = MainWnd::LISTSTATE_FIRST;

		return std::string();
	    }
	}
	
	return GetFilenameFromTreeiter(--pos_);	
	
    } else
	return GetFilenameFromTreeiter(pos_);
}

std::string Playlist::GetRandom()
{
    if(count_ != 0) {
	
	Glib::Rand rand;

	int randomNumber = rand.get_int_range(0, count_);

	pos_ = treeModel_->children().begin();
    
	for(int i = 0; i < randomNumber; i++)
	    pos_++;

	return GetFilenameFromTreeiter(pos_);
	
    } else {
	mainWnd_.listState_ = MainWnd::LISTSTATE_EMPTY;
		
	return std::string();
    }
}

void Playlist::ForceUpdatePage()
{
    while(Glib::MainContext::get_default()->iteration(false));
    OnVScrolled();
}

void Playlist::ShowCurrentFileInfoDlg()
{
    if(pos_)
	ShowFileInfoDlg( (*pos_)[columns_.filename_] );
}

bool Playlist::on_button_press_event(GdkEventButton* ev)
{
    bool rtn = TreeView::on_button_press_event(ev);

    if( (ev->type == GDK_BUTTON_PRESS) && (ev->button == 3) ) {
	menuPopup_.popup(ev->button, ev->time);
    }

    return rtn;
}

void Playlist::on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
{
    Gtk::TreeView::on_row_activated(path, column);

    if(Gtk::TreeModel::iterator iter = treeModel_->get_iter(path)) {
	pos_ = iter;
	
	mainWnd_.selectedFile_ = GetFilenameFromTreeiter(pos_);
    }

    mainWnd_.OnBtnPlay();

}

void Playlist::on_set_scroll_adjustments(Gtk::Adjustment* hadj, Gtk::Adjustment* vadj)
{
    Gtk::TreeView::on_set_scroll_adjustments(hadj, vadj);

    if(vadj) {
	vadj->signal_value_changed().connect(
	    sigc::mem_fun(*this, &Playlist::OnVScrolled) );
    }

}

void Playlist::on_size_allocate(Gtk::Allocation& allocation)
{
    Gtk::TreeView::on_size_allocate(allocation);

    // some magic constant 25
    tagProcessCount_ = allocation.get_height() / 25 + 2;
}

std::string Playlist::GetFilenameFromTreeiter(Gtk::TreeModel::iterator& iter)
{
    if(!iter) {
	// You changed reference variable!
	iter = treeModel_->children().begin();
	
	if(!iter) {
	    mainWnd_.listState_ = MainWnd::LISTSTATE_EMPTY;
	    
	    return std::string();
	}
    }

    // Mutex is locked in scope, and die at out of scope
    {
	Glib::Mutex::Lock lock(tagParsingMutex_);
	    
	if( !(*iter)[columns_.tagParsed_] ) { 
	    ParseTag(iter, false);
	}
    }

    mainWnd_.DisplayTitle( (*iter)[columns_.title_] );
    mainWnd_.DisplayTotalTime( (*iter)[columns_.time_] );
    mainWnd_.SetSeekScaleRange( (*iter)[columns_.timeSeconds_] );

    MoveCursorAndScroll(iter);

    mainWnd_.listState_ = MainWnd::LISTSTATE_NORMAL;
	
    return std::string( (*iter)[columns_.filename_] );

}

void Playlist::ParseTag(Gtk::TreeModel::iterator& iter, const bool& thread)
{
    std::string filename = (*iter)[columns_.filename_];
    
    if(IsReadable(filename)) {
	std::auto_ptr<Tag> tagParser = tagFactory_->Create(filename);

	if(tagParser.get() != NULL) { 
	    if(thread) {
		EachRow thisRow;
		
		thisRow.iter_ = iter;
		thisRow.title_ = tagParser->GetTitle();
		thisRow.time_ = tagParser->GetTotalTime();
		thisRow.timeSeconds_ = tagParser->GetTotalTimeSeconds();

		if(thisRow.timeSeconds_ != 1)
		    thisRow.tagParsed_ = true;
		
		// already locked ...
		rowQueue_.push(thisRow);

		signalRowToFill_();

	    } else {
		(*iter)[columns_.title_] = tagParser->GetTitle();
		(*iter)[columns_.time_] = tagParser->GetTotalTime();
		(*iter)[columns_.timeSeconds_] = tagParser->GetTotalTimeSeconds();

		if((*iter)[columns_.timeSeconds_] != 1)
		    (*iter)[columns_.tagParsed_] = true;

	    }
	}
    }
    
}

void Playlist::RowFill()
{
    std::auto_ptr<EachRow> processRow;
    
    {
	Glib::Mutex::Lock lock(tagParsingMutex_);

	if(rowQueue_.empty())
	    return;
	
	processRow = std::auto_ptr<EachRow>(new EachRow(rowQueue_.front()));
	rowQueue_.pop();
    }

    (*processRow->iter_)[columns_.tagParsed_] = processRow->tagParsed_;
    (*processRow->iter_)[columns_.title_] = processRow->title_;
    (*processRow->iter_)[columns_.time_] = processRow->time_;
    (*processRow->iter_)[columns_.timeSeconds_] = processRow->timeSeconds_;

}

void Playlist::GetInfoFromTag(Gtk::TreeModel::iterator& iter)
{
    unsigned int i = 0;

    while(!stopTagParsing_)
    {
	if(iter) {
	    
	    if(i > tagProcessCount_)
		break; 

	    {
		Glib::Mutex::Lock lock(tagParsingMutex_);
		
		if(!(*iter)[columns_.tagParsed_])
		    ParseTag(iter, true);	
			
	    }
	    
	    iter++;
	    i++;
	} else 
	    break;
    }
}

void Playlist::ShowFileInfoDlg(const std::string& filename)
{
    if(IsReadable(filename)) {
	std::auto_ptr<Tag> tagParser = tagFactory_->Create(filename);

	if(tagParser.get() != NULL) { 
	    std::map<std::string, Glib::ustring> fileInfoMap;
	
	    fileInfoMap["title"] = tagParser->GetTitle();
	    fileInfoMap["album"] = tagParser->GetAlbum();
	    fileInfoMap["genre"] = tagParser->GetGenre();
	    fileInfoMap["artist"] = tagParser->GetArtist();
	    fileInfoMap["comment"] = tagParser->GetComment();
	    fileInfoMap["year"] = tagParser->GetYear();
	    fileInfoMap["track"] = tagParser->GetTrack();
	
	    FileInfoDlg fileInfoDlg(fileInfoMap);
	    fileInfoDlg.run();
	}
    }
}

void Playlist::OnMenuRemove()
{
    StopThread();
    
    if(Glib::RefPtr<Gtk::TreeView::Selection> selection = get_selection()) {
	if(Gtk::TreeModel::iterator iter = selection->get_selected()) {
	    if(mainWnd_.selectedFile_ == std::string((*iter)[columns_.filename_]))
		mainWnd_.OnBtnStop();
	    
	    unsigned int prevCount = (*iter)[columns_.id_];

	    if(pos_ && pos_ == iter) 
		pos_++;
		
	    iter = treeModel_->erase(iter);	
	    MoveCursorAndScroll(iter);
	    count_--;

	    Gtk::TreeModel::iterator end = treeModel_->children().end();
	    while(iter != end) {
		(*iter)[columns_.id_] = prevCount;		
		prevCount++;
		iter++;
	    }
	}
    }
}

void Playlist::OnMenuRemoveAll()
{
    StopThread();

    mainWnd_.OnBtnStop();
    mainWnd_.selectedFile_ = "";
    mainWnd_.DisplayTitle("", true);

    Gtk::TreeModel::iterator iter = treeModel_->children().begin();

    while(iter) {
	iter = treeModel_->erase(iter);	
    }

    count_ = 0;
    pos_ = treeModel_->children().begin();
    mainWnd_.selectedFile_ = "";
}

void Playlist::OnMenuSortFilename()
{
    treeModel_->set_sort_column(columns_.filename_, Gtk::SORT_ASCENDING);

    ForceUpdatePage();
}

void Playlist::OnMenuInfo()
{
    if(Glib::RefPtr<Gtk::TreeView::Selection> selection = get_selection()) {
	if(Gtk::TreeModel::iterator iter = selection->get_selected()) {

	    ShowFileInfoDlg( (*iter)[columns_.filename_] );
	    
	}
    }

}

void Playlist::OnVScrolled()
{
    if(count_ == 0)
	return;

    StopThread();
     
    Gtk::TreePath eachPath;
    Gtk::TreeViewColumn* eachColumn;
    int cellx = 0;
    int celly = 0;

    get_path_at_pos(0, 0, eachPath, eachColumn, cellx, celly);	
    
    Gtk::TreeModel::iterator iter = treeModel_->get_iter(eachPath);
    
    inTagParsing_ = true;
    tagParsingThread_ = Glib::Thread::create( sigc::bind(
    	sigc::mem_fun(*this, &Playlist::GetInfoFromTag), iter), true );
}

void Playlist::LoadList()
{
    Tracer::Trace trace("Playlist::LoadList");
    
    std::string listFilename = Glib::get_home_dir();
    listFilename += "/.sagua/";
    listFilename += listId_;
    listFilename += ".m3u";
    
    std::ifstream listFile(listFilename.c_str());
    if(listFile) {
	std::string eachLine;
	while(getline(listFile, eachLine)) {
	    if(eachLine[0] == '#')
		continue;
	    else {
		Add(eachLine);
	    }
	}
    } else
	trace.Debug("Can't open listfile to load");
}

void Playlist::SaveList()
{
    Tracer::Trace trace("Playlist::SaveList");

    std::string programDir = Glib::get_home_dir();
    programDir += "/.sagua";

    if(!Glib::file_test(programDir, Glib::FILE_TEST_EXISTS)) {	
	trace.Debug("programDir does not exist");
	// make dir with permission 755
	mkdir(programDir.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	trace.Debug("Create programDir");
    } else if(!Glib::file_test(programDir, Glib::FILE_TEST_IS_DIR)) {
	trace.Debug("File exists already in same name");

	return;
    }

    std::string listFilename = programDir;
    listFilename += "/";
    listFilename += listId_;
    listFilename += ".m3u";	

    std::ofstream listFile(listFilename.c_str());

    if(listFile) {
	// m3u format... (is m3u a media queue format? anyway)
	//	#EXTM3U
	//	#EXTINF:song_length_in_seconds,song_title
	//	somefilepath.mp3
	listFile << "#EXTM3U" << std::endl;

	#warning "hey, expection for not parsed file?"
	Gtk::TreeModel::iterator iter = treeModel_->children().begin();

	unsigned int seconds;
	Glib::ustring title;
	std::string filename;

	while(iter) {
	    seconds = (*iter)[columns_.timeSeconds_];
	    title = (*iter)[columns_.title_];
	    filename = (*iter)[columns_.filename_];
	    listFile << "#EXTINF:" << seconds << "," << title << std::endl 
		    << filename << std::endl;
	    iter++;		    
	}
    } else
	trace.Debug("Can't open listFile to write"); 
}

void Playlist::StopThread()
{
    if(inTagParsing_) {
	stopTagParsing_ = true;
	tagParsingThread_->join();
	inTagParsing_ = false;	
	stopTagParsing_ = false;
    }
}
